#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "levit8r.h"
#include "util.h"

#define MAXSAVESIZE2 2048000
struct {
  int  lastmode, filepos, lastpos, eof;
  FILE *file;
} esclev_undo;
int esclev_undo_ready = 0;

/* ======================================================================= */
int save_esclev_now(int mode)
{
  /*
    save current gates and data file info to temp file
    mode = -1 to reset (e.g. when new file opened)
    mode =  1 to save coincidence data file info (calibs etc.)
            3 to save projections & background spectra
          + 4 to save gates
  */

  int i, m;

  if (!esclev_undo_ready) {
    esclev_undo.lastpos = -1;
    esclev_undo.lastmode = esclev_undo.filepos = esclev_undo.eof = 0;
    if (!(esclev_undo.file = tmpfile())) {
      printf("ERROR - cannot open esclev undo temp file!\n");
      exit(-1);
    }
    esclev_undo_ready = 1;
  }

  if (mode < 0) {
    esclev_undo.lastpos = -1;
    esclev_undo.lastmode = esclev_undo.filepos = esclev_undo.eof = 0;
    return 0;
  }

  if (esclev_undo.filepos > MAXSAVESIZE2) {
    rewind(esclev_undo.file);
    esclev_undo.eof = esclev_undo.filepos = 0;
  } else {
    fseek(esclev_undo.file, esclev_undo.filepos, SEEK_SET);
  }

#define W(a,b) fwrite(a, b, 1, esclev_undo.file)
  m = (mode | esclev_undo.lastmode);

  if (m&2 && xxgd.le2pro2d) {
    tell("Cannot save/undo enhanced background!\n"
	 "Resetting undo info.\n");
    save_esclev_now(-1);
    return 1;
  }
  W(&esclev_undo.lastpos, 4);
  W(&m, 4);
  if (m&1) {
    W(xxgd.luch, xxgd.numchs*4);
    W(xxgd.eff_sp, xxgd.numchs*4);
    W(xxgd.energy_sp, xxgd.numchs*4);
    W(xxgd.elo_sp, xxgd.numchs*4);
    W(xxgd.ehi_sp, xxgd.numchs*4);
    W(xxgd.ewid_sp, xxgd.numchs*4);
    W(&elgd.finest[0], 4*5);
    W(&elgd.swpars[0], 4*3);
    W(&elgd.h_norm, 4);
    W(&elgd.t_norm, 4);
    W(&elgd.q_norm, 4);
    W(&xxgd.nclook, 4);
    W(&xxgd.lookmin, 4);
    W(&xxgd.lookmax, 4);
    W(xxgd.looktab, xxgd.nclook*2);
    W(&elgd.bg_err, 4);
    W(&elgd.encal_err, 4);
    W(&elgd.effcal_err, 4);
  }
  if (m&2) {
    for (i = 0; i < 6; ++i) {
      W(xxgd.bspec[i], 4 * xxgd.numchs);
    }
    /* if (xxgd.le2pro2d) {
       W(xxgd.e2pro2d, xxgd.matchs*4);
       W(xxgd.e2e2spec, xxgd.numchs*4);
       W(&xxgd.e2e2e2sum, 4);
       } */
  }
  if (m&4) {
    for (i = 0; i < 6; ++i) {
      W(xxgd.spec[i], 4 * xxgd.numchs);
      W(xxgd.old_spec[i], 4 * xxgd.numchs);
    }
    W(&xxgd.name_gat, 80);
    W(&xxgd.old_name_gat, 80);
    for (i = 52; i < 55; ++i) {
      W(&elgd.nlist[i], 4);
      W(elgd.list[i], 4*elgd.nlist[i]);
    }
  }
#undef W

  esclev_undo.lastmode = mode;
  esclev_undo.lastpos = esclev_undo.filepos;
  esclev_undo.filepos = esclev_undo.eof = ftell(esclev_undo.file);
  return 0;
} /* save_esclev_now */

/* ======================================================================= */
int undo_esclev(int step)
{
  /* undo or redo esclev file/gate edits */
  static int extra_saved = 0;
  int  i, pos, mode;
  char dattim[20];

  if (!esclev_undo_ready || step == 0) return 1;
  if (step < 0 && (esclev_undo.lastpos < 0 ||
		   (esclev_undo.filepos > esclev_undo.eof &&
		    esclev_undo.lastpos <= esclev_undo.eof))) {
    tell("%cNo more undo information...\n", (char) 7);
    return 1;
  } else if (step > 0 && !extra_saved) {
    tell("%cNo more edits to redo...\n", (char) 7);
    return 1;
  }

  if (step < 0) {
    /* undo previous edits */
    pos = esclev_undo.lastpos;
    if (esclev_undo.filepos == esclev_undo.eof) {
      /* save current info for possible later redo */
      save_esclev_now(esclev_undo.lastmode);
    }
    extra_saved = 1;
  } else {
    /* redo previously undone edits */
    pos = esclev_undo.filepos;
    if (pos > MAXSAVESIZE2) pos = 0;
  }

  fseek(esclev_undo.file, pos, SEEK_SET);
 
#define R(a,b) fread(a, b, 1, esclev_undo.file)
  R(&esclev_undo.lastpos, 4);
  R(&mode, 4);
  /* mode = 1: coincidence data file info (calibs etc.)
            3: projections & background spectra
	  + 4: gates */
  if (mode&1) {
    R(xxgd.luch, xxgd.numchs*4);
    R(xxgd.eff_sp, xxgd.numchs*4);
    R(xxgd.energy_sp, xxgd.numchs*4);
    R(xxgd.elo_sp, xxgd.numchs*4);
    R(xxgd.ehi_sp, xxgd.numchs*4);
    R(xxgd.ewid_sp, xxgd.numchs*4);
    R(&elgd.finest[0], 4*5);
    R(&elgd.swpars[0], 4*3);
    R(&elgd.h_norm, 4);
    R(&elgd.t_norm, 4);
    R(&elgd.q_norm, 4);
    R(&xxgd.nclook, 4);
    R(&xxgd.lookmin, 4);
    R(&xxgd.lookmax, 4);
    R(xxgd.looktab, xxgd.nclook*2);
    R(&xxgd.le2pro2d, 4);
    R(&elgd.bg_err, 4);
    R(&elgd.encal_err, 4);
    R(&elgd.effcal_err, 4);

    datetime(dattim);
    if (step < 0) {
      fprintf(filez[26], "%s: Undo!\n", dattim);
    } else {
      fprintf(filez[26], "%s: Redo!\n", dattim);
    }
    fflush(filez[26]);
    read_write_l4d_file("WRITE");
  }
  if (mode&2) {
    for (i = 0; i < 6; ++i) {
      R(xxgd.bspec[i], 4 * xxgd.numchs);
      /* if (xxgd.le2pro2d) {
	 R(xxgd.e2pro2d, xxgd.matchs*4);
	 R(xxgd.e2e2spec, xxgd.numchs*4);
	 R(&xxgd.e2e2e2sum, 4);
	 } */
    }
    /* write projections etc. to .esc file */
    read_write_l4d_file("WRITE");
  }
  if (mode&4) {
    for (i = 0; i < 6; ++i) {
      R(xxgd.spec[i], 4 * xxgd.numchs);
      R(xxgd.old_spec[i], 4 * xxgd.numchs);
    }
    R(&xxgd.name_gat, 80);
    R(&xxgd.old_name_gat, 80);
    for (i = 52; i < 55; ++i) {
      R(&elgd.nlist[i], 4);
      R(elgd.list[i], 4*elgd.nlist[i]);
    }
    hilite(-1);
  }
#undef R
  esclev_undo.filepos = ftell(esclev_undo.file);

  if (step > 0 && esclev_undo.filepos == esclev_undo.eof) {
    extra_saved = 0;
    esclev_undo.filepos = pos;
  }
  calc_peaks();
  disp_dat(0);

  return 0;
} /* undo_esclev */
